/* fsm may 29 2005 s
 * june 9 2005 linesperset fix by s */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import javax.swing.*;
import javax.swing.text.JTextComponent;
import javax.swing.filechooser.FileFilter;
import javax.imageio.ImageIO;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;

public class fsm
	extends JFrame
{
	private int xchars, nsets;
	private String charset;

	private int myw, myh, myxoff, myyoff;
	private int mysize;
	private Font myfont;
	private Color[] mycolors;

	private JButton helpbutton, fontbutton, refreshbutton, colorbutton, savebutton, quitbutton;
	private JTextField wfield, hfield, xofffield, yofffield;
	private BufferedImage fontimage;
	private JLabel fontimagelabel;

	private String helptext =
		"font sheet maker will help you create\n" +
		"monospace font sheets for your favourite game.\n\n" +
		"1. choose a font (append ' bold' or '.bold' to make it bold).\n" +
		"2. adjust the offsets and refresh (optional).\n" +
		"3. change the colors (optional).\n" +
		"4. save the font.";

	public fsm()
	{
		/* default attributes */
		myw = 8;
		myh = 12;
		myxoff = 0;
		myyoff = 8;

		/* default font */
		mysize = 13;
		myfont = new Font("courier new bold", Font.BOLD, mysize);

		/* load def file */
		read_def_file("font.def");
		generate_colors();

		/* make the gui stuff */
		make_window();
		register_events();

		/* center on screen */
		setLocationRelativeTo(null);
	}

	private void make_window()
	{
		Container window;
		JPanel panel, toolbar, attributes;
		JLabel wlabel, hlabel, xofflabel, yofflabel;
		JScrollPane scrollpane;

		/* create controls */
		helpbutton = new JButton("Help");
		fontbutton = new JButton("Font");
		refreshbutton = new JButton("Refresh");
		colorbutton = new JButton("Colors");
		savebutton = new JButton("Save");
		quitbutton = new JButton("Quit");

		/* create attributes section */
		wlabel = new JLabel("Width:");
		hlabel = new JLabel("Height:");
		xofflabel = new JLabel("X offset:");
		yofflabel = new JLabel("Y offset:");

		wfield = new JTextField("" + myw, 2);
		hfield = new JTextField("" + myh, 2);
		xofffield = new JTextField("" + myxoff, 2);
		yofffield = new JTextField("" + myyoff, 2);

		attributes = new JPanel();
		attributes.setLayout(new GridLayout(2, 4, 2, 2));
		attributes.add(wlabel);
		attributes.add(wfield);
		attributes.add(hlabel);
		attributes.add(hfield);
		attributes.add(xofflabel);
		attributes.add(xofffield);
		attributes.add(yofflabel);
		attributes.add(yofffield);

		/* create default image */
		fontimage = new BufferedImage(384, 192, BufferedImage.TYPE_INT_RGB);
		ImageIcon i = new ImageIcon(fontimage);
		fontimagelabel = new JLabel(i);

		/* comment out the following line to get a 384x192 initial image */
		generate_attribs();
		update_image();

		scrollpane = new JScrollPane(fontimagelabel);

		/* create toolbar */
		toolbar = new JPanel();
		//toolbar.setLayout(new BoxLayout(toolbar, BoxLayout.X_AXIS));
		toolbar.setLayout(new FlowLayout());
		toolbar.add(helpbutton);
		toolbar.add(fontbutton);
		toolbar.add(attributes);
		toolbar.add(refreshbutton);
		toolbar.add(colorbutton);
		toolbar.add(savebutton);
		toolbar.add(quitbutton);

		/* add everything to the window */
		panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(toolbar, BorderLayout.NORTH);
		panel.add(scrollpane, BorderLayout.CENTER);

		window = getContentPane();
		window.add(panel);

		/* set title and size */
		setTitle("font sheet maker");
		pack();
	}

	private void register_events()
	{
		ActionEventHandler myactionlistener;

		/* window close event */
		setDefaultCloseOperation(EXIT_ON_CLOSE);

		/* button events */
		myactionlistener = new ActionEventHandler();
		helpbutton.addActionListener(myactionlistener);
		fontbutton.addActionListener(myactionlistener);
		refreshbutton.addActionListener(myactionlistener);
		colorbutton.addActionListener(myactionlistener);
		savebutton.addActionListener(myactionlistener);
		quitbutton.addActionListener(myactionlistener);
	}

	private void read_def_file(String filename)
	{
		try
		{
			/* open file (throws exception on failure) */
			BufferedReader in = new BufferedReader(new FileReader(filename));

			/* read lines */
			xchars = Integer.parseInt(in.readLine());
			if (xchars < 1)
				xchars = 48;

			nsets = Integer.parseInt(in.readLine());
			if (nsets < 1)
				nsets= 8;

			charset = in.readLine();
		}
		catch(Exception e)
		{
			xchars = 48;
			nsets = 8;
			charset = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNO"
				+ "PQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
		}
	}

	private void generate_colors()
	{
		int i;

		/* default colors */
		Color[] colors = new Color[8];
/*
		colors[0] = Color.LIGHT_GRAY;
		colors[1] = Color.GREEN;
		colors[2] = Color.CYAN;
		colors[3] = Color.RED;
		colors[4] = Color.YELLOW;
		colors[5] = Color.MAGENTA;
		colors[6] = Color.ORANGE;
		colors[7] = Color.PINK;
*/
		colors[0] = new Color(198, 198, 198);
		colors[1] = new Color(107, 222,  90);
		colors[2] = new Color(181, 181, 255);
		colors[3] = new Color(222,  49,   8);
		colors[4] = new Color(255, 198,  49);
		colors[5] = new Color(173, 107, 247);
		colors[6] = new Color(238, 107,  24);
		colors[7] = new Color(255, 181, 181);

		/* copy (repeatedly) to color list */
		mycolors = new Color[nsets];
		for (i = 0; i < nsets; i++)
			mycolors[i] = colors[i % 8];
	}

	private void generate_attribs()
	{
		Graphics g;
		FontMetrics fm;
		int width;

		/* load new font into gfx so we can get the metrics */
		g = fontimage.getGraphics();
		g.setFont(myfont);
		fm = g.getFontMetrics();

		/* adjust attribs using the metrics */
		myw = fm.charWidth('W');
		width = fm.charWidth('M');
		if (width > myw) myw = width;
		width = fm.charWidth('m');
		if (width > myw) myw = width;

		myh = fm.getHeight() - 3;

		myyoff = fm.getMaxAscent() - 3;
		myxoff = fm.getLeading();
	}

	/* this is the function that actually renders the font */
	private void update_image()
	{
		Graphics g;
		FontMetrics fm;
		int width, height;
		int linesperset;
		int i, j, k;

		linesperset = charset.length() / xchars;
		if (linesperset * xchars < charset.length())
			linesperset++;

		/* create image with new size */
		width = myw * xchars;
		height = myh * nsets * linesperset;
		fontimage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);

		g = fontimage.getGraphics();
		g.clearRect(0, 0, width, height);
		g.setFont(myfont);

		/* for each set */
		for (i = 0; i < nsets; i++)
		{
			g.setColor(mycolors[i]);

			/* for each line */
			for (j = 0; j < linesperset; j++)
			{
				/* for each character */
				for (k = 0; k < xchars; k++)
				{
					if (j * xchars + k < charset.length())
					{
						g.drawString("" + charset.charAt(j * xchars + k),
							myxoff + k * myw,
							myyoff + (i * linesperset + j) * myh);
					}
				}

				//g.drawString(charset.substring(j * xchars, j * xchars + xchars),
				//	0, ((i * linesperset) + (j + 1)) * myfont.getSize());
			}
		}

		/* update display */
		ImageIcon ii = new ImageIcon(fontimage);
		fontimagelabel.setIcon(ii);

		wfield.setText("" + myw);
		hfield.setText("" + myh);
		xofffield.setText("" + myxoff);
		yofffield.setText("" + myyoff);

		repaint();
	}

	private void save_font(String filename)
	{
		try
		{
			File f = new File(filename);

			/* sneakily add .png extension */
			if (!f.exists() && !filename.toLowerCase().endsWith(".png"))
			{
				filename += ".png";
				f = new File(filename);
			}

			/* warn if overwriting an existing file */
			if (f.exists())
			{
				int rv = JOptionPane.showConfirmDialog(this, f.getName()
					+ " already exists.\nDo you want to replace it?",
					"font sheet maker save",
					JOptionPane.YES_NO_OPTION,
					JOptionPane.WARNING_MESSAGE);
				if (rv == 0)
					ImageIO.write(fontimage, "png", f);
			}
			else
				ImageIO.write(fontimage, "png", f);
		}
		catch(Exception e)
		{
			JOptionPane.showMessageDialog(getInstance(),
				"An error occured when saving the image.\n"
				+ e.getMessage(),
				"font sheet maker save", JOptionPane.PLAIN_MESSAGE);
		}
	}

	/* a utility function. used to get integers out of textfields. */
	private int parse_int(JTextComponent c, int def)
	{
		int value;

		try
		{
			value = Integer.parseInt(c.getText());
		}
		catch(Exception e)
		{
			value = def;
		}

		return value;
	}

	/* might want to change this at some point in the future */
	public void show()
	{
		super.show();
	}

	/* helper function for the event handling class (see lower down) */
	private Component getInstance()
	{
		return this;
	}

	public static void main(String[] args)
	{
		fsm win = new fsm();
		win.show();
	}

	/* png file type for save dialog */
	private class MyFilter
		extends FileFilter
	{
		public boolean accept(File f)
		{
			return f.isDirectory() || f.getName().toLowerCase().endsWith(".png");
		}

		public String getDescription()
		{
			return "PNG Files (*.png)";
		}
	}

	/* the event handling class */
	private class ActionEventHandler
		implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			Object obj = e.getSource();

			if (obj == helpbutton)
			{
				JOptionPane.showMessageDialog(getInstance(), helptext,
					"font sheet maker help", JOptionPane.PLAIN_MESSAGE);
			}
			else if (obj == fontbutton)
			{
				fontchooser fc = new fontchooser(getInstance(), myfont);
				myfont = fc.getFont();

				generate_attribs();
				update_image();
			}
			else if (obj == refreshbutton)
			{
				/* read in the w/h x/yoff from the text fields */
				int i;

				i = parse_int(wfield, myw);
				if (i > 0)
					myw = i;

				i = parse_int(hfield, myh);
				if (i > 0)
					myh = i;

				myxoff = parse_int(xofffield, myxoff);
				myyoff = parse_int(yofffield, myyoff);

				update_image();
			}
			else if (obj == colorbutton)
			{
				colorchooser cc = new colorchooser(getInstance(), mycolors);
				mycolors = cc.getColors();

				update_image();
			}
			else if (obj == savebutton)
			{
				JFileChooser fc = new JFileChooser();

				FileFilter filter = new MyFilter();
				//fc.setFileFilter(filter);
				fc.addChoosableFileFilter(filter);
				fc.setAcceptAllFileFilterUsed(false);

				int rv = fc.showSaveDialog(getInstance());
				if (rv == JFileChooser.APPROVE_OPTION)
					save_font(fc.getSelectedFile().getPath());
			}
			else if (obj == quitbutton)
			{
				System.exit(0);
			}
		}
	}
}

